/* printf() */

#include <tt.h>

#define MAXIMUM_DIGITS_IN_INTEGER  10
#define HEX_10                     'A'  /* or 'a' */

static unsigned number_of_output_chars;
static unsigned minimum_field_width;
static int pad_char;
static unsigned leading_zeros;
static unsigned minus_sign;
static unsigned edited_string_length;
static boolean right_justify;

static void output_and_count(int c)
{
  standard_output(c);
  number_of_output_chars++;
}

static void output_field(char far *s)
{
  unsigned i;
  int justification_length;
  justification_length =
    minimum_field_width-leading_zeros-minus_sign-edited_string_length;
  if (right_justify)
    while (--justification_length>=0)
      output_and_count(pad_char);
  if (minus_sign) output_and_count('-');
  while (leading_zeros--) output_and_count('0');
  for (i=0; i<edited_string_length; i++)
    output_and_count(s[i]);
  while (--justification_length>=0)
    output_and_count(pad_char);
}
    

unsigned printf(char *control_string, ...)
{
  char control_char;
  int *argument = (int *)(&control_string)+1;
  number_of_output_chars = 0;
  control_char = *control_string++;
  while (control_char!='\0') 
  {
    if (control_char=='%') 
    {
      int precision = -1;
      boolean long_argument = FALSE;
      int base = 0;
      control_char = *control_string++;
      pad_char = ' ';
      minimum_field_width = 0;
      leading_zeros = 0;
      minus_sign = 0;
      right_justify = TRUE;
      if (control_char=='-') 
      {
        right_justify = FALSE;
        control_char = *control_string++;
      }
      if (control_char=='0') 
      {
        pad_char = '0';
        control_char = *control_string++;
      }
      if (control_char=='*')
        {
        minimum_field_width = *argument++;
        control_char = *control_string++;
        }
      else
        while ('0'<=control_char && control_char<='9') 
        {
          minimum_field_width = minimum_field_width*10+control_char-'0';
          control_char = *control_string++;
        }
      if (control_char=='.') 
      {
        precision = 0;
        while ('0'<=(control_char=*control_string++) && control_char<='9')
          precision = precision*10+control_char-'0';
      }
      if (control_char=='l') 
      {
        long_argument = TRUE;
        control_char = *control_string++;
      }
      if (control_char=='d')
        base = 10;
      else if (control_char=='x') 
        base = 16;
      else if (control_char=='u') 
        base = 10;
      else if (control_char=='o') 
        base = 8;
      else if (control_char=='c')
      {
        base = -1;
        pad_char = ' ';
      }
      else if (control_char=='s')
      {
        base = -2;
        pad_char = ' ';
      }
      if (base==0)  /* invalid conversion type */
      {
        if (control_char!='\0') {
          output_and_count(control_char);
          control_char = *control_string++;
          }
      }
      else
      {
        if (base==(-1))  /* conversion type c */
        {
          edited_string_length = 1;
          output_field((char *)(argument++));
        }
        else if (base==(-2))  /* conversion type s */
        {
          char far *string;
          edited_string_length = 0;
          if (long_argument)
          {
            string = * (char far **) argument;
            argument += sizeof(char far *)/sizeof(int);
          }
          else
          {
            string = * (char **) argument;
            argument += sizeof(char *)/sizeof(int);
          }
          while (string[edited_string_length]!=0)
            edited_string_length++;
          if (precision>=0 && edited_string_length>precision)
            edited_string_length = precision;
          output_field(string);
        }
        else  /* conversion type d, o or x */
        {
          unsigned long x;
          char buffer[MAXIMUM_DIGITS_IN_INTEGER+1];
          edited_string_length = 0;
          if (long_argument) 
          {
            x = * (unsigned long *) argument;
            argument += sizeof(unsigned long)/sizeof(int);
          }
          else if (control_char=='d')
            x = (long) *argument++;
          else
            x = (unsigned) *argument++;
          if (control_char=='d' && (long) x < 0) 
          {
            minus_sign = 1;
            x = - (long) x;
          }
          do 
          {
            unsigned c;
            c = divide_and_remainder(&x,base) + '0';
            if (c>'9')
              c += HEX_10-'9'-1;
            buffer[sizeof(buffer)-1-edited_string_length++] = c;
          }
          while (x!=0);
          if (precision>=0 && precision>edited_string_length)
            leading_zeros = precision-edited_string_length;
          output_field(buffer+sizeof(buffer)-edited_string_length);
        }
        control_char = *control_string++;
      }
    }
    else
    {
      output_and_count(control_char);
      control_char = *control_string++;
    }
  }
  return number_of_output_chars;
}
