/* open_serial(), close_serial(), receive_byte() */

#pragma inline
#include <tt.h>

/* I/O register offsets from base address */

#define RECEIVED_DATA             0
#define INTERRUPT_IDENTIFICATION  2
#define LINE_STATUS               5
#define MODEM_STATUS              6
#define TRANSMITTED_DATA          0
#define INTERRUPT_ENABLE          1
#define LS_BAUD_COUNT             0
#define MS_BAUD_COUNT             1
#define LINE_CONTROL              3
#define MODEM_CONTROL             4

extern void _serial_interrupt();

struct serial *open_serial(unsigned base, unsigned irq, unsigned baud,
  unsigned parity, unsigned data, unsigned stop, unsigned size)
{
  register struct serial *s = malloc(sizeof(struct serial));
  if (s==NULL) return s;
  s->buffer = malloc(size);
   if (s->buffer == NULL) {free(s); return NULL;}
  s->base = base;
  s->irq = irq;
  s->index = s->count = 0;
  s->size = size-1;
  s->mask = (1<<data)-1;
  s->overflow = FALSE;
  s->interrupt_entry.push_si = 0x56;
  s->interrupt_entry.move_si = 0xBE;
  s->interrupt_entry.pointer = s;
  s->interrupt_entry.jmp = 0xE9;
  s->interrupt_entry.handler =
    (unsigned) _serial_interrupt - (unsigned)(&s->interrupt_entry.handler+1);
  __outportb__(base+MODEM_CONTROL, 0); /* stop interrupts */
  {
    unsigned intnum = irq + (irq<8?8:112);
    s->save_interrupt_vector = interrupt_vector(intnum);
    set_interrupt_vector(intnum, (void far *)(&s->interrupt_entry));
  }
  {
    unsigned parity_stop_data = parity<<3|(stop-1)<<2|(data-5);
    unsigned pic = irq<8 ? 0x21 : 0xA1;
    unsigned mask = ~(1<<(irq&7));
    _DX = 115200>>16;
    _AX = 115200;
    asm div word ptr baud;
    _BX = _AX;
    __cli__();
    __outportb__(base+LINE_CONTROL, 0x80);
    __outportb__(base+LS_BAUD_COUNT, _BL);
    __outportb__(base+MS_BAUD_COUNT, _BH);
    __outportb__(base+LINE_CONTROL, parity_stop_data);
    __outportb__(base+MODEM_CONTROL, 8+3);
    __outportb__(base+INTERRUPT_ENABLE, 1);
    __outportb__(pic, __inportb__(pic)&mask);
    __inportb__(base);
    __sti__();
  }
  return s;
}

void close_serial(struct serial *s)
{
  unsigned pic = s->irq<8 ? 0x21 : 0xA1;
  __outportb__(s->base+MODEM_CONTROL, 0); /* stop interrupts */
  set_interrupt_vector(s->irq + (s->irq<8?8:112), s->save_interrupt_vector);
  __outportb__(pic, __inportb__(pic) & ~(1<<(s->irq&7)));
  free(s->buffer);
  free(s);
}

int receive_byte(struct serial *s)
{
  int r = -1;
  if (s->count != 0)
  {
    __cli__();
    r = s->buffer[(s->index - s->count--) & s->size];
    __sti__();
    r &= s->mask;
  }
  return r;
}

